// Png2Img.cpp : Defines the entry point for the console application.
//
// JBUM - added 16-bit dither support - use -d option.  SLOW!

#include "stdafx.h"

#include <stdlib.h>

#ifndef PNG_DEBUG
#define PNG_DEBUG 0
#endif

#include "png.h"

typedef unsigned char byte;
typedef unsigned char ui8;
typedef unsigned long ui32;

////////////////////////////////////////////////////////////////////////////////
// ".IMG" file types supported:
//
#define fModePtv16    0x0001
#define fModePtv16a   0x0101
#define fModePtvP8    0x0100


int gDither = 0, gPatDither = 0;
int Dither16(int r, int g, int b, int x, int y);
static double RandomD(void);

////////////////////////////////////////////////////////////////////////////////
//#include "remdebug.h"
//
#ifndef D
#define D if(0)
#endif
#ifndef DP
#define DP(ARGS) D{printf##ARGS; printf("\n");}
#endif
#ifndef FATAL
#define FATAL(ARGS)                                \
{                                                  \
printf( "FATAL[%s:%d]: ", __FILE__, __LINE__ );    \
printf##ARGS;                                      \
printf("\n");                                      \
exit(1);                                           \
}
#endif
#ifndef WARN
#define WARN(ARGS)                                 \
{                                                  \
printf( "WARNING[%s:%d]: ", __FILE__, __LINE__ );  \
printf##ARGS;                                      \
printf("\n");                                      \
rc=1;                                              \
}
#endif

static int rc = 0;


typedef FILE * FileHandle;
#define FileRead( FILE, DATA, LENGTH ) fread( DATA, 1, LENGTH, FILE )
#define FileWrite( FILE, DATA, LENGTH ) fwrite( DATA, 1, LENGTH, FILE )

////////////////////////////////////////////////////////////////////////////////
//#include "remfile.h"
//

#define PutUI32( F, V ) PutUI( F, V, 4 )
#define PutUI24( F, V ) PutUI( F, V, 3 )
#define PutUI16( F, V ) PutUI( F, V, 2 )
#define PutUI8( F, V ) PutUI( F, V, 1 )

#define Abs(X) (((X) < 0) ? -(X) : (X))
#define LimitUnsigned( X, MAX ) ( (X)>(MAX) ? (MAX) : (X) )
#define LimitPixel( X ) LimitUnsigned( X, 255 )


typedef enum P8PlaneType
   {
   kP8PlaneNode,                                // collection of sub-planes
   kP8PlaneGray,                                // monochrome image
   kP8PlaneRGBi,                                // RGB indexed color
   kP8PlaneAlpha,                               // alpha
   kP8PlaneNx,                                  // x normal
   kP8PlaneNy,                                  // y normal
   kP8PlaneNz,                                  // z normal
   kP8PlaneRefl,                                // reflectance index
   kP8PlaneR,                                   // red
   kP8PlaneG,                                   // green
   kP8PlaneB,                                   // blue
   kP8PlaneUnknown                              // ...skip it
   } P8PlaneType;


////////////////////////////////////////////////////////////////////////////////
static void DisassembleNumber( byte * data, unsigned long value, int l_data )
// Convert this machines number to a byte string
// if l_data < 0, the least significant byte is first
////////////////////////////////////////////////////////////////////////////////
   {
   const int abs_l_data = Abs( l_data );

   if( l_data > 0 )
      {
      for( int i = abs_l_data-1 ; i >= 0 ; i-- )
         {
         data[i] = (byte) (value & 0xff);
         value = (value >> 8);
         }
      }
   else
      {
      for( int i = 0 ; i < abs_l_data ; i++ )
         {
         data[i] = (byte) (value & 0xff);
         value = (value >> 8);
         }
      }
   }


////////////////////////////////////////////////////////////////////////////////
static void PutBytes( FILE * f, unsigned long value, int l_data )
// if l_data < 0, the least significant byte is first
////////////////////////////////////////////////////////////////////////////////
   {
   const int abs_l_data = Abs( l_data );
   int rc;
   byte * data = new byte[ abs_l_data ];

   DisassembleNumber( data, value, l_data );

   // Put the bytes containing the data
   rc = fwrite( data, sizeof( byte ), abs_l_data, f );
   if( rc != abs_l_data ) FATAL(( "fwrite: rc=%d", rc ));
   delete data;
   }


///////////////////////////////////////////////////////////////////////////////
static void PutUI( FILE * f, unsigned long value, int l_data )
// Put an n-byte value to f (n specified by l_data), checking to make sure that
// it is within range, and do it in a byte-order-independent fashion.
///////////////////////////////////////////////////////////////////////////////
   {
   const int abs_l_data = Abs( l_data );
   const int max_l_data = sizeof( unsigned long ) / sizeof( byte );
   const unsigned long max_value = ( (unsigned long)(-1) >> ( 8 * (4-abs_l_data) ));

   // Check for programming errors
   if( value > max_value ) FATAL(( "PutUI(f,v,l) called with v=%ld > max=%ld",
                                   value, max_value ));

   if( -max_l_data > l_data
       ||
       l_data > max_l_data ) FATAL(( "PutUI(f,l) called with l < -%d or l > %d",
                                     max_l_data, max_l_data ));

   PutBytes( f, value, l_data );
   }


//
////////////////////////////////////////////////////////////////////////////////



#if defined(PNG_TIME_RFC1123_SUPPORTED)
static int tIME_chunk_present=0;
static char tIME_string[30] = "no tIME chunk present in file";
#endif /* PNG_TIME_RFC1123_SUPPORTED */

int convert_one_file PNGARG((PNG_CONST char *inname, PNG_CONST char *outname));

#ifdef __TURBOC__
#include <mem.h>
#endif

/* defined so I can write to a file on gui/windowing platforms */
/*  #define STDERR stderr  */
#define STDERR stdout   /* for DOS */

/* example of using row callbacks to make a simple progress meter */
static int status_pass=1;
static int status_dots_requested=0;
static int status_dots=1;



void
read_row_callback(png_structp png_ptr, png_uint_32 row_number, int pass)
{
    if(png_ptr == NULL || row_number > 0x3fffffffL) return;
    if(status_pass != pass)
    {
       D fprintf(stdout,"\n Pass %d: ",pass);
       status_pass = pass;
       status_dots = 30;
    }
    status_dots--;
    if(status_dots == 0)
    {
       D fprintf(stdout, "\n         ");
       status_dots=30;
    }
    D fprintf(stdout, "r");
}



void
write_row_callback(png_structp png_ptr, png_uint_32 row_number, int pass)
{
    if(png_ptr == NULL || row_number > 0x3fffffffL || pass > 7) return;
    D fprintf(stdout, "w");
}



#if defined(PNG_WRITE_USER_TRANSFORM_SUPPORTED)
/* example of using user transform callback (we don't transform anything,
   but merely count the zero samples) */

static png_uint_32 zero_samples;

void
count_zero_samples(png_structp png_ptr, png_row_infop row_info, png_bytep data)
{
   png_bytep dp = data;
   if(png_ptr == NULL)return;

   /* contents of row_info:
    *  png_uint_32 width      width of row
    *  png_uint_32 rowbytes   number of bytes in row
    *  png_byte color_type    color type of pixels
    *  png_byte bit_depth     bit depth of samples
    *  png_byte channels      number of channels (1-4)
    *  png_byte pixel_depth   bits per pixel (depth*channels)
    */

    /* counts the number of zero samples (or zero pixels if color_type is 3 */

    if(row_info->color_type == 0 || row_info->color_type == 3)
    {
       int pos=0;
       png_uint_32 n, nstop;
       for (n=0, nstop=row_info->width; n<nstop; n++)
       {
          if(row_info->bit_depth == 1)
          {
             if(((*dp << pos++ )& 0x80) == 0) zero_samples++;
             if(pos == 8)
             {
                pos = 0;
                dp++;
             }
          }
          if(row_info->bit_depth == 2)
          {
             if(((*dp << (pos+=2))& 0xc0) == 0) zero_samples++;
             if(pos == 8)
             {
                pos = 0;
                dp++;
             }
          }
          if(row_info->bit_depth == 4)
          {
             if(((*dp << (pos+=4))& 0xf0) == 0) zero_samples++;
             if(pos == 8)
             {
                pos = 0;
                dp++;
             }
          }
          if(row_info->bit_depth == 8)
             if(*dp++ == 0) zero_samples++;
          if(row_info->bit_depth == 16)
          {
             if((*dp | *(dp+1)) == 0) zero_samples++;
             dp+=2;
          }
       }
    }
    else /* other color types */
    {
       png_uint_32 n, nstop;
       int channel;
       int color_channels = row_info->channels;
       if(row_info->color_type > 3)color_channels--;

       for (n=0, nstop=row_info->width; n<nstop; n++)
       {
          for (channel = 0; channel < color_channels; channel++)
          {
             if(row_info->bit_depth == 8)
                if(*dp++ == 0) zero_samples++;
             if(row_info->bit_depth == 16)
             {
                if((*dp | *(dp+1)) == 0) zero_samples++;
                dp+=2;
             }
          }
          if(row_info->color_type > 3)
          {
             dp++;
             if(row_info->bit_depth == 16)dp++;
          }
       }
    }
}
#endif /* PNG_WRITE_USER_TRANSFORM_SUPPORTED */



static int verbose = 0;
//static int wrote_question = 0;

#if defined(PNG_NO_STDIO)
/* START of code to validate stdio-free compilation */
/* These copies of the default read/write functions come from pngrio.c and */
/* pngwio.c.  They allow "don't include stdio" testing of the library. */
/* This is the function that does the actual reading of data.  If you are
   not reading from a standard C stream, you should create a replacement
   read_data function and use it at run time with png_set_read_fn(), rather
   than changing the library. */
#ifndef USE_FAR_KEYWORD
static void
png_default_read_data(png_structp png_ptr, png_bytep data, png_size_t length)
{
   png_size_t check;

   /* fread() returns 0 on error, so it is OK to store this in a png_size_t
    * instead of an int, which is what fread() actually returns.
    */
   check = (png_size_t)fread(data, (png_size_t)1, length,
      (FILE *)png_ptr->io_ptr);

   if (check != length)
   {
      png_error(png_ptr, "Read Error");
   }
}
#else
/* this is the model-independent version. Since the standard I/O library
   can't handle far buffers in the medium and small models, we have to copy
   the data.
*/

#define NEAR_BUF_SIZE 1024
#define MIN(a,b) (a <= b ? a : b)



static void
png_default_read_data(png_structp png_ptr, png_bytep data, png_size_t length)
{
   int check;
   png_byte *n_data;
   FILE *io_ptr;

   /* Check if data really is near. If so, use usual code. */
   n_data = (png_byte *)CVT_PTR_NOCHECK(data);
   io_ptr = (FILE *)CVT_PTR(png_ptr->io_ptr);
   if ((png_bytep)n_data == data)
   {
      check = fread(n_data, 1, length, io_ptr);
   }
   else
   {
      png_byte buf[NEAR_BUF_SIZE];
      png_size_t read, remaining, err;
      check = 0;
      remaining = length;
      do
      {
         read = MIN(NEAR_BUF_SIZE, remaining);
         err = fread(buf, (png_size_t)1, read, io_ptr);
         png_memcpy(data, buf, read); /* copy far buffer to near buffer */
         if(err != read)
            break;
         else
            check += err;
         data += read;
         remaining -= read;
      }
      while (remaining != 0);
   }
   if (check != length)
   {
      png_error(png_ptr, "read Error");
   }
}
#endif /* USE_FAR_KEYWORD */



#if defined(PNG_WRITE_FLUSH_SUPPORTED)
static void
png_default_flush(png_structp png_ptr)
{
   FILE *io_ptr;
   io_ptr = (FILE *)CVT_PTR((png_ptr->io_ptr));
   if (io_ptr != NULL)
      fflush(io_ptr);
}
#endif

/* This is the function that does the actual writing of data.  If you are
   not writing to a standard C stream, you should create a replacement
   write_data function and use it at run time with png_set_write_fn(), rather
   than changing the library. */
#ifndef USE_FAR_KEYWORD
static void
png_default_write_data(png_structp png_ptr, png_bytep data, png_size_t length)
{
   png_uint_32 check;

   check = fwrite(data, 1, length, (FILE *)(png_ptr->io_ptr));
   if (check != length)
   {
      png_error(png_ptr, "Write Error");
   }
}
#else
/* this is the model-independent version. Since the standard I/O library
   can't handle far buffers in the medium and small models, we have to copy
   the data.
*/

#define NEAR_BUF_SIZE 1024
#define MIN(a,b) (a <= b ? a : b)



static void
png_default_write_data(png_structp png_ptr, png_bytep data, png_size_t length)
{
   png_uint_32 check;
   png_byte *near_data;  /* Needs to be "png_byte *" instead of "png_bytep" */
   FILE *io_ptr;

   /* Check if data really is near. If so, use usual code. */
   near_data = (png_byte *)CVT_PTR_NOCHECK(data);
   io_ptr = (FILE *)CVT_PTR(png_ptr->io_ptr);
   if ((png_bytep)near_data == data)
   {
      check = fwrite(near_data, 1, length, io_ptr);
   }
   else
   {
      png_byte buf[NEAR_BUF_SIZE];
      png_size_t written, remaining, err;
      check = 0;
      remaining = length;
      do
      {
         written = MIN(NEAR_BUF_SIZE, remaining);
         png_memcpy(buf, data, written); /* copy far buffer to near buffer */
         err = fwrite(buf, 1, written, io_ptr);
         if (err != written)
            break;
         else
            check += err;
         data += written;
         remaining -= written;
      }
      while (remaining != 0);
   }
   if (check != length)
   {
      png_error(png_ptr, "Write Error");
   }
}

#endif /* USE_FAR_KEYWORD */



/* This function is called when there is a warning, but the library thinks
 * it can continue anyway.  Replacement functions don't have to do anything
 * here if you don't want to.  In the default configuration, png_ptr is
 * not used, but it is passed in case it may be useful.
 */
static void
png_default_warning(png_structp png_ptr, png_const_charp message)
{
   PNG_CONST char *name = "UNKNOWN (ERROR!)";
   if (png_ptr != NULL && png_ptr->error_ptr != NULL)
      name = png_ptr->error_ptr;
   fprintf(STDERR, "%s: libpng warning: %s\n", name, message);
}

/* This is the default error handling function.  Note that replacements for
 * this function MUST NOT RETURN, or the program will likely crash.  This
 * function is used by default, or if the program supplies NULL for the
 * error function pointer in png_set_error_fn().
 */
static void
png_default_error(png_structp png_ptr, png_const_charp message)
{
   png_default_warning(png_ptr, message);
   /* We can return because png_error calls the default handler, which is
    * actually OK in this case. */
}
#endif /* PNG_NO_STDIO */
/* END of code to validate stdio-free compilation */



/* START of code to validate memory allocation and deallocation */
#ifdef PNG_USER_MEM_SUPPORTED

/* Allocate memory.  For reasonable files, size should never exceed
   64K.  However, zlib may allocate more then 64K if you don't tell
   it not to.  See zconf.h and png.h for more information.  zlib does
   need to allocate exactly 64K, so whatever you call here must
   have the ability to do that.

   This piece of code can be compiled to validate max 64K allocations
   by setting MAXSEG_64K in zlib zconf.h *or* PNG_MAX_MALLOC_64K. */
typedef struct memory_information {
   png_uint_32                    size;
   png_voidp                 pointer;
   struct memory_information FAR *next;
} memory_information;
typedef memory_information FAR *memory_infop;

static memory_infop pinformation = NULL;
static int current_allocation = 0;
static int maximum_allocation = 0;

extern PNG_EXPORT(png_voidp,png_debug_malloc) PNGARG((png_structp png_ptr,
   png_uint_32 size));
extern PNG_EXPORT(void,png_debug_free) PNGARG((png_structp png_ptr,
   png_voidp ptr));



png_voidp
png_debug_malloc(png_structp png_ptr, png_uint_32 size) {

   /* png_malloc has already tested for NULL; png_create_struct calls
      png_debug_malloc directly, with png_ptr == NULL which is OK */

   if (size == 0)
      return (png_voidp)(NULL);

   /* This calls the library allocator twice, once to get the requested
      buffer and once to get a new free list entry. */
   {
      memory_infop pinfo = png_malloc_default(png_ptr, sizeof *pinfo);
      pinfo->size = size;
      current_allocation += size;
      if (current_allocation > maximum_allocation)
         maximum_allocation = current_allocation;
      pinfo->pointer = png_malloc_default(png_ptr, size);
      pinfo->next = pinformation;
      pinformation = pinfo;
      /* Make sure the caller isn't assuming zeroed memory. */
      png_memset(pinfo->pointer, 0xdd, pinfo->size);
      return (png_voidp)(pinfo->pointer);
   }
}



/* Free a pointer.  It is removed from the list at the same time. */
void
png_debug_free(png_structp png_ptr, png_voidp ptr)
{
   if (png_ptr == NULL)
      fprintf(STDERR, "NULL pointer to png_debug_free.\n");
   if (ptr == 0) {
#if 0 /* This happens all the time. */
      fprintf(STDERR, "WARNING: freeing NULL pointer\n");
#endif
      return;
   }

   /* Unlink the element from the list. */
   {
      memory_infop FAR *ppinfo = &pinformation;
      for (;;) {
         memory_infop pinfo = *ppinfo;
         if (pinfo->pointer == ptr) {
            *ppinfo = pinfo->next;
            current_allocation -= pinfo->size;
            if (current_allocation < 0)
               fprintf(STDERR, "Duplicate free of memory\n");
            /* We must free the list element too, but first kill
               the memory that is to be freed. */
            memset(ptr, 0x55, pinfo->size);
            png_free_default(png_ptr, pinfo);
            break;
         }
         if (pinfo->next == NULL) {
            fprintf(STDERR, "Pointer %x not found\n", ptr);
            break;
         }
         ppinfo = &pinfo->next;
      }
   }

   /* Finally free the data. */
   png_free_default(png_ptr, ptr);
}
#endif /* PNG_USER_MEM_SUPPORTED */
/* END of code to test memory allocation/deallocation */



/* Test one file */
int
convert_one_file(PNG_CONST char *inname, PNG_CONST char *outname)
   {
   static FILE *fpin, *fpout;  /* "static" prevents setjmp corruption */
   png_structp read_ptr /*, write_ptr*/ ;
   png_infop read_info_ptr /*, write_info_ptr*/ , end_info_ptr;
   png_bytep row_buf;
   png_uint_32 row, col;
   //png_uint_32 y;
   png_uint_32 width, height;
   //int num_pass, pass;
   int bit_depth, color_type;
#ifdef USE_FAR_KEYWORD
   jmp_buf jmpbuf;
#endif

   //char inbuf[256], outbuf[256];

   row_buf = (png_bytep)NULL;

   if ((fpin = fopen(inname, "rb")) == NULL)
      {
      fprintf(STDERR, "Could not find input file %s\n", inname);
      return (1);
      }

   if ((fpout = fopen(outname, "wb")) == NULL)
      {
      fprintf(STDERR, "Could not open output file %s\n", outname);
      fclose(fpin);
      return (1);
      }

   png_debug(0, "Allocating read and write structures\n");
#ifdef PNG_USER_MEM_SUPPORTED
   read_ptr = png_create_read_struct_2(
      PNG_LIBPNG_VER_STRING, (png_voidp)NULL,
      (png_error_ptr)NULL, (png_error_ptr)NULL, (png_voidp)NULL,
      (png_malloc_ptr)png_debug_malloc, (png_free_ptr)png_debug_free);
#else
   read_ptr = png_create_read_struct(
      PNG_LIBPNG_VER_STRING, (png_voidp)NULL,
      (png_error_ptr)NULL, (png_error_ptr)NULL);
#endif
#if defined(PNG_NO_STDIO)
   png_set_error_fn(read_ptr, (png_voidp)inname, png_default_error,
                    png_default_warning);
#endif

// #ifdef PNG_USER_MEM_SUPPORTED
//    write_ptr = png_create_write_struct_2(PNG_LIBPNG_VER_STRING, (png_voidp)NULL,
//       (png_error_ptr)NULL, (png_error_ptr)NULL, (png_voidp)NULL,
//       (png_malloc_ptr)png_debug_malloc, (png_free_ptr)png_debug_free);
// #else
//    write_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, (png_voidp)NULL,
//       (png_error_ptr)NULL, (png_error_ptr)NULL);
// #endif

// #if defined(PNG_NO_STDIO)
//    png_set_error_fn(write_ptr, (png_voidp)inname, png_default_error,
//        png_default_warning);
// #endif
   png_debug(0, "Allocating read_info, write_info and end_info structures\n");
   read_info_ptr = png_create_info_struct(read_ptr);
//    write_info_ptr = png_create_info_struct(write_ptr);
   end_info_ptr = png_create_info_struct(read_ptr);
#ifdef PNG_USER_MEM_SUPPORTED
#endif

   png_debug(0, "Setting jmpbuf for read struct\n");
#ifdef USE_FAR_KEYWORD
   if (setjmp(jmpbuf))
#else
   if (setjmp(read_ptr->jmpbuf))
#endif
      {
      fprintf(STDERR, "%s -> %s: libpng read error\n", inname, outname);
      png_destroy_read_struct(&read_ptr, &read_info_ptr, &end_info_ptr);
//       png_destroy_write_struct(&write_ptr, &write_info_ptr);
      fclose(fpin);
      fclose(fpout);
      return (1);
      }
#ifdef USE_FAR_KEYWORD
   png_memcpy(read_ptr->jmpbuf,jmpbuf,sizeof(jmp_buf));
#endif

//    png_debug(0, "Setting jmpbuf for write struct\n");
// #ifdef USE_FAR_KEYWORD
//    if (setjmp(jmpbuf))
// #else
//    if (setjmp(write_ptr->jmpbuf))
// #endif
//    {
//       fprintf(STDERR, "%s -> %s: libpng write error\n", inname, outname);
//       png_destroy_read_struct(&read_ptr, &read_info_ptr, &end_info_ptr);
//       png_destroy_write_struct(&write_ptr, &write_info_ptr);
//       fclose(fpin);
//       fclose(fpout);
//       return (1);
//    }
// #ifdef USE_FAR_KEYWORD
//    png_memcpy(write_ptr->jmpbuf,jmpbuf,sizeof(jmp_buf));
// #endif

   png_debug(0, "Initializing input and output streams\n");
#if !defined(PNG_NO_STDIO)
   png_init_io(read_ptr, fpin);

//    png_init_io(write_ptr, fpout);

#else
   png_set_read_fn(read_ptr, (png_voidp)fpin, png_default_read_data);
//    png_set_write_fn(write_ptr, (png_voidp)fpout,  png_default_write_data,
// #if defined(PNG_WRITE_FLUSH_SUPPORTED)
//       png_default_flush);
// #else
//       NULL);
// #endif
#endif
   if(status_dots_requested == 1)
      {
//       png_set_write_status_fn(write_ptr, write_row_callback);
      png_set_read_status_fn(read_ptr, read_row_callback);
      }
   else
      {
//       png_set_write_status_fn(write_ptr, NULL);
      png_set_read_status_fn(read_ptr, NULL);
      }

#  if defined(PNG_WRITE_USER_TRANSFORM_SUPPORTED)
   zero_samples=0;
//      png_set_write_user_transform_fn(write_ptr, count_zero_samples);
#  endif

   png_debug(0, "Reading info struct\n");
   png_read_info(read_ptr, read_info_ptr);

   png_debug(0, "Transferring info struct\n");
      {
      int interlace_type, compression_type, filter_type;

      if (png_get_IHDR(read_ptr, read_info_ptr, &width, &height, &bit_depth,
                       &color_type, &interlace_type, &compression_type, &filter_type))
         {
//          png_set_IHDR(write_ptr, write_info_ptr, width, height, bit_depth,
// #if defined(PNG_WRITE_INTERLACING_SUPPORTED)
//             color_type, interlace_type, compression_type, filter_type);
// #else
//             color_type, PNG_INTERLACE_NONE, compression_type, filter_type);
// #endif
         if( color_type != PNG_COLOR_TYPE_RGB_ALPHA
             &&
             color_type != PNG_COLOR_TYPE_RGB
             &&
             color_type != PNG_COLOR_TYPE_GRAY )
            {
            FATAL(( "Input file color type must be RGBA, RGB or P8.GRAY" ));
            }
         }
      }
#if defined(PNG_READ_bKGD_SUPPORTED) && defined(PNG_WRITE_bKGD_SUPPORTED)
      {
      png_color_16p background;

      if (png_get_bKGD(read_ptr, read_info_ptr, &background))
         {
//          png_set_bKGD(write_ptr, write_info_ptr, background);
         }
      }
#endif
#if defined(PNG_READ_cHRM_SUPPORTED) && defined(PNG_WRITE_cHRM_SUPPORTED)
      {
      double white_x, white_y, red_x, red_y, green_x, green_y, blue_x, blue_y;

      if (png_get_cHRM(read_ptr, read_info_ptr, &white_x, &white_y, &red_x,
                       &red_y, &green_x, &green_y, &blue_x, &blue_y))
         {
//          png_set_cHRM(write_ptr, write_info_ptr, white_x, white_y, red_x,
//             red_y, green_x, green_y, blue_x, blue_y);
         }
      }
#endif
#if defined(PNG_READ_gAMA_SUPPORTED) && defined(PNG_WRITE_gAMA_SUPPORTED)
      {
      double gamma;

      if (png_get_gAMA(read_ptr, read_info_ptr, &gamma))
         {
//          png_set_gAMA(write_ptr, write_info_ptr, gamma);
         }
      }
#endif
#if defined(PNG_READ_sRGB_SUPPORTED) && defined(PNG_WRITE_sRGB_SUPPORTED)
      {
      int intent;

      if (png_get_sRGB(read_ptr, read_info_ptr, &intent))
         {
//          png_set_sRGB(write_ptr, write_info_ptr, intent);
         }
      }
#endif
#if defined(PNG_READ_hIST_SUPPORTED) && defined(PNG_WRITE_hIST_SUPPORTED)
      {
      png_uint_16p hist;

      if (png_get_hIST(read_ptr, read_info_ptr, &hist))
         {
//          png_set_hIST(write_ptr, write_info_ptr, hist);
         }
      }
#endif
#if defined(PNG_READ_oFFs_SUPPORTED) && defined(PNG_WRITE_oFFs_SUPPORTED)
      {
      png_int_32 offset_x, offset_y;
      int unit_type;

      if (png_get_oFFs(read_ptr, read_info_ptr, &offset_x,&offset_y,&unit_type))
         {
//          png_set_oFFs(write_ptr, write_info_ptr, offset_x, offset_y, unit_type);
         }
      }
#endif
#if defined(PNG_READ_pCAL_SUPPORTED) && defined(PNG_WRITE_pCAL_SUPPORTED)
      {
      png_charp purpose, units;
      png_charpp params;
      png_int_32 X0, X1;
      int type, nparams;

      if (png_get_pCAL(read_ptr, read_info_ptr, &purpose, &X0, &X1, &type,
                       &nparams, &units, &params))
         {
//          png_set_pCAL(write_ptr, write_info_ptr, purpose, X0, X1, type,
//             nparams, units, params);
         }
      }
#endif
#if defined(PNG_READ_pHYs_SUPPORTED) && defined(PNG_WRITE_pHYs_SUPPORTED)
      {
      png_uint_32 res_x, res_y;
      int unit_type;

      if (png_get_pHYs(read_ptr, read_info_ptr, &res_x, &res_y, &unit_type))
         {
//          png_set_pHYs(write_ptr, write_info_ptr, res_x, res_y, unit_type);
         }
      }
#endif
      {
      png_colorp palette;
      int num_palette;

      if (png_get_PLTE(read_ptr, read_info_ptr, &palette, &num_palette))
         {
//          png_set_PLTE(write_ptr, write_info_ptr, palette, num_palette);
         }
      }
#if defined(PNG_READ_sBIT_SUPPORTED) && defined(PNG_WRITE_sBIT_SUPPORTED)
      {
      png_color_8p sig_bit;

      if (png_get_sBIT(read_ptr, read_info_ptr, &sig_bit))
         {
//          png_set_sBIT(write_ptr, write_info_ptr, sig_bit);
         }
      }
#endif
#if (defined(PNG_READ_tEXt_SUPPORTED) && defined(PNG_WRITE_tEXt_SUPPORTED)) || \
    (defined(PNG_READ_zTXt_SUPPORTED) && defined(PNG_WRITE_zTXt_SUPPORTED))
      {
      png_textp text_ptr;
      int num_text;

      if (png_get_text(read_ptr, read_info_ptr, &text_ptr, &num_text) > 0)
         {
         png_debug1(0, "Handling %d tEXt/zTXt chunks\n", num_text);
//          png_set_text(write_ptr, write_info_ptr, text_ptr, num_text);
         }
      }
#endif
#if defined(PNG_READ_tIME_SUPPORTED) && defined(PNG_WRITE_tIME_SUPPORTED)
      {
      png_timep mod_time;

      if (png_get_tIME(read_ptr, read_info_ptr, &mod_time))
         {
//          png_set_tIME(write_ptr, write_info_ptr, mod_time);
#if defined(PNG_TIME_RFC1123_SUPPORTED)
         /* we have to use png_strcpy instead of "=" because the string
            pointed to by png_convert_to_rfc1123() gets free'ed before
            we use it */
         png_strcpy(tIME_string,png_convert_to_rfc1123(read_ptr, mod_time));
         tIME_chunk_present++;
#endif /* PNG_TIME_RFC1123_SUPPORTED */
         }
      }
#endif
#if defined(PNG_READ_tRNS_SUPPORTED) && defined(PNG_WRITE_tRNS_SUPPORTED)
      {
      png_bytep trans;
      int num_trans;
      png_color_16p trans_values;

      if (png_get_tRNS(read_ptr, read_info_ptr, &trans, &num_trans,
                       &trans_values))
         {
//          png_set_tRNS(write_ptr, write_info_ptr, trans, num_trans,
//             trans_values);
         }
      }
#endif

   png_debug(0, "\nWriting info struct\n");
//    png_write_info(write_ptr, write_info_ptr);

   png_debug(0, "\nAllocating row buffer \n");
   row_buf = (png_bytep)png_malloc(read_ptr,
      png_get_rowbytes(read_ptr, read_info_ptr));
   if (row_buf == NULL)
      {
      fprintf(STDERR, "No memory to allocate row buffer\n");
      png_destroy_read_struct(&read_ptr, &read_info_ptr, (png_infopp)NULL);
//       png_destroy_write_struct(&write_ptr, &write_info_ptr);
      fclose(fpin);
      fclose(fpout);
      return (1);
      }
   png_debug(0, "Writing row data\n");

   // allocate all the row buffers
   png_bytep * image = new png_bytep[ height ];
   for( row = 0 ; row < height ; row++ )
      {
      image[row] = (png_bytep)png_malloc(read_ptr,
         png_get_rowbytes(read_ptr, read_info_ptr));
      }
   png_read_image( read_ptr, image );

   switch( color_type )
      {
      case PNG_COLOR_TYPE_RGB_ALPHA:
         {
         // Put header
         PutUI16( fpout, fModePtv16a );
         PutUI16( fpout, width );
         PutUI16( fpout, height );
         PutUI16( fpout, 3*width );

         D printf( "Output .IMG file: fMode=0x%04lx, width=0x%04lx, "
                   "height=0x%04lx, rowbytes=0x%04lx \n",
                   fModePtv16a, width, height, 3*width );

         // Build 16+8-bit pixels
         ui32 pixel;
         for( row = 0 ; row < height ; row++ )
            {
            for( col = 0 ; col < width*4 ; col+=4 )
               {
								// Only dither if pixel is fully opaque
								if (gDither && image[row][col+3] == 0x0ff)
									pixel = ((ui32)LimitPixel( image[row][col+3] ) & 0xff ) << (16) |
										      Dither16(image[row][col+0],image[row][col+1],image[row][col+2],col/4,row);
								else
								 pixel =
										((ui32)LimitPixel( image[row][col+3] ) & 0xff ) << (16) | //a
										((ui32)LimitPixel( image[row][col+0] ) & 0xf8 ) << (8) |  //r
										((ui32)LimitPixel( image[row][col+1] ) & 0xfc ) << (3) |  //g
										((ui32)LimitPixel( image[row][col+2] ) & 0xf8 ) >> (3);   //b
								 PutUI24( fpout, pixel );
               }
            }
         break;
         }

      case PNG_COLOR_TYPE_RGB:
         {
         // Put header
         PutUI16( fpout, fModePtv16 );
         PutUI16( fpout, width );
         PutUI16( fpout, height );
         PutUI16( fpout, 2*width );

         D printf( "Output .IMG file: fMode=0x%04lx, width=0x%04lx, "
                   "height=0x%04lx, rowbytes=0x%04lx \n",
                   fModePtv16a, width, height, 3*width );

         // Build 16-bit pixels
         ui32 pixel;
         for( row = 0 ; row < height ; row++ )
            {
            for( col = 0 ; col < width*3 ; col+=3 )
               {
								if (gDither)
									pixel = Dither16(image[row][col+0],image[row][col+1],image[row][col+2],col/3,row);
								else
	               pixel =
                  ((ui32)LimitPixel( image[row][col+0] ) & 0xf8 ) << (8) |
                  ((ui32)LimitPixel( image[row][col+1] ) & 0xfc ) << (3) |
                  ((ui32)LimitPixel( image[row][col+2] ) & 0xf8 ) >> (3);
               PutUI16( fpout, pixel );
               }
            }
         break;
         }

      case PNG_COLOR_TYPE_GRAY:
         {
         // Put header
         PutUI16( fpout, fModePtvP8 );
         PutUI16( fpout, width );
         PutUI16( fpout, height );
         PutUI16( fpout, 1*width );  //rowBytes

         D printf( "Output .IMG file: fMode=0x%04lx, width=0x%04lx, "
                   "height=0x%04lx, rowbytes=0x%04lx \n",
                   fModePtvP8, width, height, 1*width );

         // write additional P8 header information
         PutUI16( fpout, 1 );  //nPlanes
         FileWrite( fpout, "GRAY", 4 );  //pMode
         PutUI32( fpout, 1+(width*height) );  //pLength
         PutUI8( fpout, 0 );  //pName

         // Write 8-bit pixels
         ui32 pixel;
         for( row = 0 ; row < height ; row++ )
            {
            for( col = 0 ; col < width*1 ; col+=1 )
               {
               pixel = image[row][col+0];
               PutUI8( fpout, LimitPixel( pixel ));
               }
            }
         break;
         }
      }

//    num_pass = png_set_interlace_handling(read_ptr);
// //    png_set_interlace_handling(write_ptr);
//
//    for (pass = 0; pass < num_pass; pass++)
//    {
//       png_debug1(0, "Writing row data for pass %d\n",pass);
//       for (y = 0; y < height; y++)
//       {
//          png_read_rows(read_ptr, (png_bytepp)&row_buf, (png_bytepp)NULL, 1);
// //          png_write_rows(write_ptr, (png_bytepp)&row_buf, 1);
//       }
//    }

   png_debug(0, "Reading and writing end_info data\n");
   png_read_end(read_ptr, end_info_ptr);
//    png_write_end(write_ptr, end_info_ptr);

// #ifdef PNG_EASY_ACCESS_SUPPORTED
//    if(verbose)
//    {
//       png_uint_32 iwidth, iheight;
//       iwidth = png_get_image_width(write_ptr, write_info_ptr);
//       iheight = png_get_image_height(write_ptr, write_info_ptr);
//       fprintf(STDERR, "Image width = %lu, height = %lu\n",
//          iwidth, iheight);
//    }
// #endif

   png_debug(0, "Destroying data structs\n");
   delete image;
   png_free(read_ptr, row_buf);
   png_destroy_read_struct(&read_ptr, &read_info_ptr, &end_info_ptr);
//    png_destroy_write_struct(&write_ptr, &write_info_ptr);

   fclose(fpin);
   fclose(fpout);

// /////////////////////////////////////////////////////////////////////////////
//
//    png_debug(0, "Opening files for comparison\n");
//    if ((fpin = fopen(inname, "rb")) == NULL)
//    {
//       fprintf(STDERR, "Could not find file %s\n", inname);
//       return (1);
//    }
//
//    if ((fpout = fopen(outname, "rb")) == NULL)
//    {
//       fprintf(STDERR, "Could not find file %s\n", outname);
//       fclose(fpin);
//       return (1);
//    }
//
//    for(;;)
//    {
//       png_size_t num_in, num_out;
//
//       num_in = fread(inbuf, 1, 1, fpin);
//       num_out = fread(outbuf, 1, 1, fpout);
//
//       if (num_in != num_out)
//       {
//          fprintf(STDERR, "Files %s and %s are of a different size\n",
//                  inname, outname);
//          if(wrote_question == 0)
//          {
//             fprintf(STDERR,
//               "   Was %s written with the same chunk size (8k),",inname);
//             fprintf(STDERR,
//               " filtering\n   heuristic (libpng default), compression");
//             fprintf(STDERR,
//               " level (zlib default)\n   and zlib version (%s)?\n\n",
//               ZLIB_VERSION);
//             wrote_question=1;
//          }
//          fclose(fpin);
//          fclose(fpout);
//          return (0);
//       }
//
//       if (!num_in)
//          break;
//
//       if (png_memcmp(inbuf, outbuf, num_in))
//       {
//          fprintf(STDERR, "Files %s and %s are different\n", inname, outname);
//          if(wrote_question == 0)
//          {
//             fprintf(STDERR,
//               "   Was %s written with the same chunk size (8k),",inname);
//             fprintf(STDERR,
//               " filtering\n   heuristic (libpng default), compression");
//             fprintf(STDERR,
//               " level (zlib default)\n   and zlib version (%s)?\n\n",
//               ZLIB_VERSION);
//             wrote_question=1;
//          }
//          fclose(fpin);
//          fclose(fpout);
//          return (0);
//       }
//    }
//
//    fclose(fpin);
//    fclose(fpout);
//
   return (0);
   }


void SyntaxError()
{
  fprintf(STDERR,
             "\nUsage: png2img <in.png> <out.img>\n\nOptions:\n\t-d Dither (RGB888 -> RGB565)\n\t-p Pattern Dither\n\t-v Verbose\n\t-m Multiple Input Files\n");
 	exit(1);
}

/* input and output filenames */
// #ifdef RISCOS
// PNG_CONST char *inname = "pngtest/png";
// PNG_CONST char *outname = "pngout/png";
// #else
static PNG_CONST char *inname = NULL;
static PNG_CONST char *outname = NULL;
static char outbuf[256];
// #endif

char *genOutputName(const char *instr)
{
	strcpy(outbuf, instr);
	char *p = strchr(outbuf,'.');
	char isUpper = 0;
	if (p != NULL) {
		isUpper = (p[1] == 'P');
		*p = '\0';
	}
	strcat(outbuf,isUpper? ".IMG" : ".img");
	return outbuf;
}


int
main(int argc, char *argv[])
{
   int ierror = 0, firstFile=0,  multiple=0;

   D fprintf(STDERR, "Testing libpng version %s\n", PNG_LIBPNG_VER_STRING);
   D fprintf(STDERR, "   with zlib   version %s\n", ZLIB_VERSION);
   D fprintf(STDERR,"%s",png_get_copyright(NULL));

   /* Do some consistency checking on the memory allocation settings, I'm
      not sure this matters, but it is nice to know, the first of these
      tests should be impossible because of the way the macros are set
      in pngconf.h */
#if defined(MAXSEG_64K) && !defined(PNG_MAX_MALLOC_64K)
      fprintf(STDERR, " NOTE: Zlib compiled for max 64k, libpng not\n");
#endif
   /* I think the following can happen. */
#if !defined(MAXSEG_64K) && defined(PNG_MAX_MALLOC_64K)
      fprintf(STDERR, " NOTE: libpng compiled for max 64k, zlib not\n");
#endif

	int i;

		// rewrote parser to simplify option handling - jbum
	 for (i = 1; i < argc; ++i)
	 {
		 if (argv[i][0] == '-') {
				for (char *p = &argv[i][1]; *p; ++p) {
					switch(tolower(*p)) {
					case 'v':
						verbose = 1;
						status_dots_requested = 1;
						break;
					case 'm':
						multiple = 1;	// only strictly needed if two input files
													// otherwise, we figure it out
						break;
					case 'd':
						gDither = 1;
						break;
					case 'p':
						gDither = 1;
						gPatDither = 1;
						break;
					default:
						SyntaxError();
					}
			 }
		 }
		 else {
			 if (firstFile == 0)
				 firstFile = i;
			 if (outname)
			 	 multiple = 1;
			 else if (inname)
         outname = argv[i];
			 else
         inname = argv[i];
		 }
	 }

	 if (inname == NULL)
	 	SyntaxError();

	 // Use input name to generate output name
	 if (outname == NULL) {
		 outname = genOutputName(inname);
	 }


	 int kerror = 0;
#ifdef PNG_USER_MEM_SUPPORTED
	 int allocation_now = current_allocation;
#endif
	if (multiple) {
			for (int i = firstFile; i < argc && !kerror; ++i) {
				inname = argv[i];
				outname = genOutputName(inname);
				if (verbose)
					fprintf(STDERR, "Converting %s -> %s\n",inname, outname);
				kerror = convert_one_file(inname, outname);
			}
	}
	else {
	 if (verbose)
			fprintf(STDERR, "Converting %s -> %s\n",inname, outname);
	 kerror = convert_one_file(inname, outname);
	}

  return (int)(rc != 0 || kerror != 0);
}

long 	gSeed = 1;
#define	R_A	16807L
#define	R_M	2147483647L
#define R_Q	127773L
#define R_R	2836L
static long RandomL(void);
static void Randomize(void);
static void MySRand(long s);
static long GetSRand(void);

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

static long GetSRand(void)
{
	return gSeed;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

static void MySRand(long s)
{
	gSeed = s;
	if (gSeed == 0)
		gSeed = 1;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

static void Randomize(void)
{
	MySRand(time(NULL));
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

static long RandomL(void)
{
	long	hi,lo,test;

	hi   = gSeed / R_Q;
	lo   = gSeed % R_Q;
	test = R_A * lo - R_R * hi;
	if (test > 0)
    gSeed = test;
	else
	  gSeed = test + R_M;
	return gSeed;
}

static double RandomD(void)
{
	return RandomL() / (double) R_M;
}

int Dither16(int r, int g, int b, int x, int y)
{
	static int sBayerDither[] = {0,8,2,10,12,4,14,6,3,11,1,9,15,7,13,5};

	if (gPatDither)
	{
		int patAddr = ((y & 3)<<2)+(x & 3);
		r = (r & 0xf8) + (sBayerDither[patAddr] < ((r & 7)<<1)?  0x08 : 0);
		g = (g & 0xfc) + (sBayerDither[patAddr] < ((g & 3)<<2)?  0x04 : 0);
		b = (b & 0xf8) + (sBayerDither[patAddr] < ((b & 7)<<1)?  0x08 : 0);
	}
	else {	// Random Dither
		r = (r & 0xf8) + ( RandomD() < (r & 7)/8.0? 0x08 : 0);
		g = (g & 0xfc) + ( RandomD() < (g & 3)/4.0? 0x04 : 0);
		b = (b & 0xf8) + ( RandomD() < (b & 7)/8.0? 0x08 : 0 );
	}
	r = (r > 255? 255 : r) & 0xf8;
	g = (g > 255? 255 : g) & 0xfc;
	b = (b > 255? 255 : b) & 0xf8;
	return (r << 8) | (g << 3) | (b >> 3);
}
