#!/usr/bin/perl
use strict;

open (OFILE, ">parthenon.bzw") || die ("Can't open platforms for output");
print OFILE <<EOT;
# Parthenon
#
# by His Blind Ambition  (jbum\@jbum.com)
#
#
#
EOT

my ($nx,$ny) = (8,17);
my $coldist = 20.0;  # 20.0
my $colthick = $coldist*0.45;
my $coltall = $colthick*5;
my $baseheight = $colthick;
my $captall = $colthick/3.0;
my $capwide = $colthick/4.0;
my $pwidth = $coldist*($nx-1);
my $pdepth = $coldist*($ny-1);
my $pwrad = $pwidth/2.0;
my $pdrad = $pdepth/2.0;
my $capbase = $baseheight+$coltall+$colthick;

MakeParth(20,0,0,0);
# make mini-parth on table
my $legh = 0.3;
MakeParth(0.4,0,0,$baseheight+$legh+.1);
PrintMazeBox(0,0,$baseheight+$legh,3.7,3.7,.1,0);
PrintMazeBox(-3.4,-3.4,$baseheight,.1,.1,$legh,0);
PrintMazeBox(-3.4, 3.4,$baseheight,.1,.1,$legh,0);
PrintMazeBox( 3.4,-3.4,$baseheight,.1,.1,$legh,0);
PrintMazeBox( 3.4, 3.4,$baseheight,.1,.1,$legh,0);

# sentinel
MakeTeleporter(0,-$colthick,0,$coldist/2,$coltall,90,0);  # 0,1
MakeTeleporter(0, $colthick,0,$coldist/2,$coltall,90,0);	# 2,3
MakeTeleporter($coldist/2,0,0,$colthick,$coltall, 0,0);		# 4,5
MakeTeleporter(-$coldist/2,0,0,$colthick,$coltall,0,0);		# 6,7

# teleporter on roof
MakeTeleporter(0, 0, $capbase, $colthick, $coltall/2, 0, 0);					# 8,9
# teleporters on corners
MakeTeleporter(340,340,30,10,30,45,0);	  # 10,11
MakeTeleporter(-340,-340,30,10,30,45,0);	# 12,13
MakeTeleporter(340,-340,30,10,30,-45,0);	# 14,15
MakeTeleporter(-340,340,30,10,30,-45,0);	# 16,17
# interior walls
my $wallthick = $colthick/2;	# 18,19,20,21,22,23,24,25
MakeTeleporter(-($pwrad-$coldist*1),-$wallthick,$baseheight,
 								$pdrad-($coldist*2+$wallthick),$coltall,0,0);
MakeTeleporter( ($pwrad-$coldist*1),$wallthick,$baseheight,
 								$pdrad-($coldist*2+$wallthick),$coltall,0,0);
MakeTeleporter(-$wallthick*2,-($pdrad-$coldist*2),$baseheight,
 								$pwrad-($coldist*1+$wallthick*2),,$coltall,90,0);
MakeTeleporter($wallthick*2, ($pdrad-$coldist*2),$baseheight,
 								$pwrad-($coldist*1+$wallthick*2),$coltall,90,0);

foreach (0..3)
{
	MakeLink($_*2, 8);
	MakeLink($_*2+1, 9);
}
# setup top links and corners to make continuous circuit with lots of diagonal crosses
MakeLink(8,11);	
MakeLink(9,12);	
MakeLink(10,14);
MakeLink(11,13); 
MakeLink(14,16);	
MakeLink(12,10);	
MakeLink(13,15); 
MakeLink(15,17); 
MakeLink(16,8);
MakeLink(17,9);

foreach(18..25)
{
	MakeLink($_,$_);
}


#random trees
foreach (0..150)
{
	my ($tx,$ty);
	while (1) {
		($tx,$ty) = (int(rand()*750)-375,int(rand()*750)-375);
		last if ($tx < -($pwrad+$colthick*3) || $tx > $pwrad+$colthick*3 ||
				$ty < -($pdrad+$colthick*3) || $ty > $pdrad+$colthick*3 ||
				$tx < -380 || $tx > 380 ||
				$ty < -380 || $ty > 380);
	}
	my $theight = 10+int(rand()*20);
	my $tbase = $theight*(.1 + rand()*.3);
	my ($tthick) = $theight*.06;
	my ($tbush) = $tthick+rand()*($theight-$tthick)*.75;
	PrintMazeBox($tx,$ty,0,$tthick/2.0,$tthick/2.0,$tbase,rand()*90);
	PrintMazePyr($tx,$ty,$tbase,$tbush/2.0,$tbush/2.0,$theight-$tbase,rand()*90);
}

# include overworld

sub MakeParth($$$$)
{
	my ($coldist,$ox,$oy,$oz) = @_;
	my $colthick = $coldist*0.45;
	my $baseheight = $colthick;
	my $coltall = $colthick*5;
	my $captall = $colthick/3.0;
	my $capwide = $colthick/4.0;
	my $pwidth = $coldist*($nx-1);
	my $pdepth = $coldist*($ny-1);
	my $pwrad = $pwidth/2.0;
	my $pdrad = $pdepth/2.0;

	foreach my $x (0..$nx-1)
	{
		PrintColumn($ox+$coldist*$x-$pwrad,$oy+$pdrad,$coltall,$oz,$coldist);
		PrintColumn($ox+$coldist*$x-$pwrad,$oy-$pdrad,$coltall,$oz,$coldist);
	}

		foreach my $x (1..$nx-2)
		{
			PrintColumn($ox+$coldist*$x-$pwrad,$oy+$pdrad-$coldist,$coltall,$oz,$coldist);
			PrintColumn($ox+$coldist*$x-$pwrad,$oy-($pdrad-$coldist),$coltall,$oz,$coldist);
		}


		foreach my $y (1..$ny-2)
		{
			PrintColumn($ox+$pwidth/2,$ox+$coldist*$y-$pdrad,$coltall,$oz,$coldist);
			PrintColumn($ox-$pwidth/2,$oy+$coldist*$y-$pdrad,$coltall,$oz,$coldist);
		}

		# Lower Step
		my $swid = $colthick*2;
		my $xrad = $pwrad+$colthick*2;
		my $yrad = $pdrad+$colthick*2;
		PrintMazeBox($ox-$xrad,$oy,$oz,$swid/2,$yrad+$swid/2,$baseheight/2.0,0);
		PrintMazeBox($ox+$xrad,$oy,$oz,$swid/2,$yrad+$swid/2,$baseheight/2.0,0);
		PrintMazeBox($ox,$oy+$yrad,$oz,$xrad-$swid/2,$swid/2,$baseheight/2.0,0);
		PrintMazeBox($ox,$oy-$yrad,$oz,$xrad-$swid/2,$swid/2,$baseheight/2.0,0);

		# Upper Step
		PrintMazeBox($ox,$oy,$oz+$baseheight/2,
									$pwrad+$colthick,$pdrad+$colthick,$baseheight/2,0);


		# Roof
		PrintMazeBox($ox,$oy,$oz+$baseheight+$coltall,
									$pwrad+$colthick,$pdrad+$colthick,$colthick/2,0);
		PrintMazeBox($ox,$oy,$oz+$baseheight+$coltall+$colthick/2.0,
									$pwrad+$colthick*1.5,$pdrad+$colthick*1.5,$colthick/2,0);

		my $capbase = $baseheight+$coltall+$colthick;

		# Roof Caps
		PrintMazePyr($ox, $oy+$pdrad, $oz+$capbase, $pwrad+$colthick*1.5, $colthick, $coltall/2);
		PrintMazePyr($ox, $oy-$pdrad, $oz+$capbase, $pwrad+$colthick*1.5, $colthick, $coltall/2);
		PrintMazePyr($ox, $oy,        $oz+$capbase, $pwrad+$colthick*1.5, $colthick, $coltall/2);
}


sub PrintColumn($$$$$)
{
	my ($x,$y,$tall,$z,$coldist) = @_;

	my $colthick = $coldist*0.45;
	my $baseheight = $colthick;
	my $coltall = $colthick*5;
	my $captall = $colthick/3.0;
	my $capwide = $colthick/4.0;

	if ($coldist > 1) {
		PrintMazeBox($x,$y,$z,$colthick/2,$colthick/4,$baseheight+$coltall-$captall,0);
		PrintMazeBox($x,$y,$z,$colthick/2,$colthick/4,$baseheight+$coltall-$captall,60);
		PrintMazeBox($x,$y,$z,$colthick/2,$colthick/4,$baseheight+$coltall-$captall,-60);
	}
	else {
		PrintMazeBox($x,$y,$z,$colthick/2,$colthick/4,$baseheight+$coltall-$captall,45);
	}
#	PrintMazeBox($x,$y,$z,$colthick/2,$colthick/4,$baseheight+$coltall-$captall,90);
	# cap
	PrintMazeBox($x,$y,$z+$baseheight+$coltall-$captall,
								$colthick/2+$capwide,$colthick/2+$capwide,$captall,0);
}

sub PrintMazeBox($$$$$$)
{
	my ($x,$y,$z,$xs,$ys,$zs, $r) = @_;
	printf OFILE "box\nposition %6.3f %6.3f %6.3f\nrotation %6.3f\nsize %6.3f %6.3f %6.3f\nend\n\n",
							$x,$y,$z,$r,$xs,$ys,$zs;
}

sub PrintMazePyr($$$$$$)
{
	my ($x,$y,$z,$xs,$ys,$zs, $r) = @_;

	printf OFILE "pyramid\nposition %6.3f %6.3f %6.3f\nrotation %6.3f\nsize %6.3f %6.3f %6.3f\nend\n\n",
							$x,$y,$z,$r,$xs,$ys,$zs;
}

sub MakeTeleporter($$$$$$)
{
	my ($x,$y,$z,$w,$h,$rot,$b) = @_;
	printf OFILE "teleporter\nposition %6.3f %6.3f %6.3f\nrotation %6.3f\nsize 0 %6.3f %6.3f\nborder %d\nend\n\n",
							$x,$y,$z,$rot,$w,$h,$b;
}

sub MakeLink($$)
{
	my ($from,$to) = @_;
	print OFILE <<EOT;
link
from $from
to   $to
end
EOT
}
