#!/usr/bin/perl
open (OFILE, ">quad.bzw") || die ("Can't open output file\n");
print OFILE <<EOT;
# IKEA Quadrophenia
#
# by His Blind Ambition  (jbum\@jbum.com)
#
#
#
EOT

my $tw = 200;
my $ntiles = int(800/$tw);
foreach my $y (0..$ntiles-1)
{
	my $oy = $tw*.4+$y*$tw-400;
	foreach my $x (0..$ntiles-1)
	{
		my $ox = $tw*.4+$x*$tw-400;
		MakeSpikes($ox,$oy,120);
		MakeTile($ox,$oy,11,1);
		MakeTile($ox,$oy,22,.8);
		MakeTile($ox,$oy,33,.6);
		MakeTile($ox,$oy,44,.4);
	}
}


#teleporter on roof
MakeTeleporter(10,10,300,10,30,45,1);	  # 0,1

#teleporter on bottom
MakeTeleporter(0,0,0,10,30,45,1);	      # 2,3

# teleporters on corners
MakeTeleporter(380,380,0,10,30,45,1);	  # 4,5
MakeTeleporter(-380,-380,0,10,30,45,1);	# 6,7
MakeTeleporter(380,-380,0,10,30,-45,1);	# 8,9
MakeTeleporter(-380,380,0,10,30,-45,1);	# 10,11

MakeLink(0,3);
MakeLink(1,2);
MakeLink(2,0);
MakeLink(3,1);
MakeLink(4,0);
MakeLink(5,6);
MakeLink(6,2);
MakeLink(7,4);
MakeLink(8,1);
MakeLink(9,10);
MakeLink(10,3);
MakeLink(11,8);

close OFILE;
exit;


sub MakeTile()
{
	my ($ox,$oy,$h,$sc) = @_;

	my $sw = $tw/5;
	my $dw = sqrt(2.0)*$sw;
	my $fw = 1.0;  # floor width
	PrintBox($ox+$tw/2,$oy,$h,$sw,$sc*$sw/2,$fw,0);	
	PrintBox($ox,$oy+$tw/2,$h,$sc*$sw/2,$sw,$fw,0);
	if (rand() < .5)
	{
		PrintBox($ox+$tw/5,$oy+$tw/5,$h,$dw,$sc*$dw/2,$fw,-45);
		PrintBox($ox-$tw/5,$oy-$tw/5,$h,$dw,$sc*$dw/2,$fw,-45);
	}
	else {
		PrintBox($ox-$tw/5,$oy+$tw/5,$h,$dw,$sc*$dw/2,$fw,45);
		PrintBox($ox+$tw/5,$oy-$tw/5,$h,$dw,$sc*$dw/2,$fw,45);
	}
}

sub MakeSpikes()
{
	my ($ox,$oy,$h) = @_;
#	MakeSpike($ox-$tw*1.5/5, $oy, 0, $tw/20,$h);
#	MakeSpike($ox+$tw*1.5/5, $oy, 0, $tw/20,$h);
#	MakeSpike($ox,$oy-$tw*1.5/5,  0, $tw/20,$h);
#	MakeSpike($ox,$oy+$tw*1.5/5,  0, $tw/20,$h);

	MakeSpike($ox-$tw/5, $oy-$tw/5, 0, $tw/20,$h);
	MakeSpike($ox-$tw/5, $oy+$tw/5, 0, $tw/20,$h);
	MakeSpike($ox+$tw/5, $oy-$tw/5, 0, $tw/20,$h);
	MakeSpike($ox+$tw/5, $oy+$tw/5, 0, $tw/20,$h);

}


sub MakeSpike()
{
	my ($x,$y,$z,$w,$h) = @_;
	PrintPyramid($x,$y,$z,$w/2,$w/2,$h,rand()*360);
	# PrintPyramid($x,$y,$z+$h,$w/2,$w/2,10,rand()*360);
}

sub MakeColumn($$$$$)
{
	my ($x,$y,$z,$colthick,$coltall,$r) = @_;

	my $baseheight = 0;
	my $captall = $colthick/5.0;
	my $capwide = $coltall/3.0;

	PrintBox($x,$y,$z,$colthick/2,$colthick/4,$baseheight+$coltall-$captall,$r+0);
	PrintBox($x,$y,$z,$colthick/2,$colthick/4,$baseheight+$coltall-$captall,$r+60);
	PrintBox($x,$y,$z,$colthick/2,$colthick/4,$baseheight+$coltall-$captall,$r-60);
	PrintBox($x,$y,$z+$baseheight+$coltall-$captall,
								$colthick/2+$capwide,$colthick/2+$capwide,$captall,$r);
}

sub PrintBox()
{
	my ($x,$y,$z,$xs,$ys,$zs,$r) = @_;
	printf OFILE "box\n";
	printf OFILE "position %7.3f %7.3f %7.3f\n",$x,$y,$z;
	printf OFILE "rotation %7.3f\n",$r;
	printf OFILE "size %7.3f %7.3f %7.3f\n",$xs,$ys,$zs;
	printf OFILE "end\n\n";
}

sub PrintPyramid($$$$$$)
{
	my ($x,$y,$z,$xs,$ys,$zs, $r) = @_;

	printf OFILE "pyramid\nposition %6.3f %6.3f %6.3f\nrotation %6.3f\nsize %6.3f %6.3f %6.3f\nend\n\n",
							$x,$y,$z,$r,$xs,$ys,$zs;
}

sub MakeTeleporter($$$$$$)
{
	my ($x,$y,$z,$w,$h,$rot,$b) = @_;
	printf OFILE "teleporter\nposition %6.3f %6.3f %6.3f\nrotation %6.3f\nsize 0 %6.3f %6.3f\nborder %d\nend\n\n",
							$x,$y,$z,$rot,$w,$h,$b;
}



sub MakeLink($$)
{
	my ($from,$to) = @_;
	printf OFILE "link\nfrom %d\nto %d\nend\n\n",$from,$to;
}
