//#define REMDEBUG
////////////////////////////////////////////////////////////////////////////////
//
// png2bzw - convert a PNG image to PTV ".IMG" format file
//
////////////////////////////////////////////////////////////////////////////////

#include <stdafx.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

/* Makes pngtest verbose so we can find problems (needs to be before png.h) */
#ifndef PNG_DEBUG
#define PNG_DEBUG 0
#endif

#include "png.h"


typedef unsigned char byte;
typedef unsigned char ui8;
typedef unsigned long ui32;

static int rc = 0;


typedef FILE * FileHandle;
#define FileRead( FILE, DATA, LENGTH ) fread( DATA, 1, LENGTH, FILE )
#define FileWrite( FILE, DATA, LENGTH ) fwrite( DATA, 1, LENGTH, FILE )

////////////////////////////////////////////////////////////////////////////////
//#include "remfile.h"
//


#define Abs(X) (((X) < 0) ? -(X) : (X))
#define LimitUnsigned( X, MAX ) ( (X)>(MAX) ? (MAX) : (X) )
#define LimitPixel( X ) LimitUnsigned( X, 255 )


typedef enum P8PlaneType
   {
   kP8PlaneNode,                                // collection of sub-planes
   kP8PlaneGray,                                // monochrome image
   kP8PlaneRGBi,                                // RGB indexed color
   kP8PlaneAlpha,                               // alpha
   kP8PlaneNx,                                  // x normal
   kP8PlaneNy,                                  // y normal
   kP8PlaneNz,                                  // z normal
   kP8PlaneRefl,                                // reflectance index
   kP8PlaneR,                                   // red
   kP8PlaneG,                                   // green
   kP8PlaneB,                                   // blue
   kP8PlaneUnknown                              // ...skip it
   } P8PlaneType;


typedef struct {
	png_bytep		*image;
	png_bytep		*hmap;
	png_uint_32	width,height;
	int					pixelSize; // in bytes
} ImageRef;

int MeasureWidth(int col, int row, ImageRef *img, int pix);
int MeasureHeight(int col, int row, ImageRef *img, int pix);
void MarkBlock(int col, int row, int w, int h, int match, ImageRef *img);
int FindHighest(int col, int row, int pw, int ph, ImageRef *img);
void MeasureLargestBlock(int col, int row, int match, ImageRef *img, int *pw, int *ph);
int GetLowestPixel(int col, int row, int pw, int ph, ImageRef *img);

////////////////////////////////////////////////////////////////////////////////
static void DisassembleNumber( byte * data, unsigned long value, int l_data )
// Convert this machines number to a byte string
// if l_data < 0, the least significant byte is first
////////////////////////////////////////////////////////////////////////////////
   {
   const int abs_l_data = Abs( l_data );

   if( l_data > 0 )
      {
      for( int i = abs_l_data-1 ; i >= 0 ; i-- )
         {
         data[i] = (byte) (value & 0xff);
         value = (value >> 8);
         }
      }
   else
      {
      for( int i = 0 ; i < abs_l_data ; i++ )
         {
         data[i] = (byte) (value & 0xff);
         value = (value >> 8);
         }
      }
   }


//
////////////////////////////////////////////////////////////////////////////////

#if defined(PNG_TIME_RFC1123_SUPPORTED)
static int tIME_chunk_present=0;
static char tIME_string[30] = "no tIME chunk present in file";
#endif /* PNG_TIME_RFC1123_SUPPORTED */

int convert_one_file PNGARG((PNG_CONST char *inname, PNG_CONST char *outname));

#ifdef __TURBOC__
#include <mem.h>
#endif

/* defined so I can write to a file on gui/windowing platforms */
/*  #define STDERR stderr  */
#define STDERR stdout   /* for DOS */

static int gVerbose = 0, gInvert = 0, gClip = 0, gLandscape;
double	gFloorOfst, gMapHeight=256, gMapWidth=800, gPicWidth, gPicHeight, gTall;

//static int wrote_question = 0;

#if defined(PNG_NO_STDIO)
/* START of code to validate stdio-free compilation */
/* These copies of the default read/write functions come from pngrio.c and */
/* pngwio.c.  They allow "don't include stdio" testing of the library. */
/* This is the function that does the actual reading of data.  If you are
   not reading from a standard C stream, you should create a replacement
   read_data function and use it at run time with png_set_read_fn(), rather
   than changing the library. */
#ifndef USE_FAR_KEYWORD
static void
png_default_read_data(png_structp png_ptr, png_bytep data, png_size_t length)
{
   png_size_t check;

   /* fread() returns 0 on error, so it is OK to store this in a png_size_t
    * instead of an int, which is what fread() actually returns.
    */
   check = (png_size_t)fread(data, (png_size_t)1, length,
      (FILE *)png_ptr->io_ptr);

   if (check != length)
   {
      png_error(png_ptr, "Read Error");
   }
}
#else
/* this is the model-independent version. Since the standard I/O library
   can't handle far buffers in the medium and small models, we have to copy
   the data.
*/

#define NEAR_BUF_SIZE 1024
#define MIN(a,b) (a <= b ? a : b)



static void
png_default_read_data(png_structp png_ptr, png_bytep data, png_size_t length)
{
   int check;
   png_byte *n_data;
   FILE *io_ptr;

   /* Check if data really is near. If so, use usual code. */
   n_data = (png_byte *)CVT_PTR_NOCHECK(data);
   io_ptr = (FILE *)CVT_PTR(png_ptr->io_ptr);
   if ((png_bytep)n_data == data)
   {
      check = fread(n_data, 1, length, io_ptr);
   }
   else
   {
      png_byte buf[NEAR_BUF_SIZE];
      png_size_t read, remaining, err;
      check = 0;
      remaining = length;
      do
      {
         read = MIN(NEAR_BUF_SIZE, remaining);
         err = fread(buf, (png_size_t)1, read, io_ptr);
         png_memcpy(data, buf, read); /* copy far buffer to near buffer */
         if(err != read)
            break;
         else
            check += err;
         data += read;
         remaining -= read;
      }
      while (remaining != 0);
   }
   if (check != length)
   {
      png_error(png_ptr, "read Error");
   }
}
#endif /* USE_FAR_KEYWORD */



#if defined(PNG_WRITE_FLUSH_SUPPORTED)
static void
png_default_flush(png_structp png_ptr)
{
   FILE *io_ptr;
   io_ptr = (FILE *)CVT_PTR((png_ptr->io_ptr));
   if (io_ptr != NULL)
      fflush(io_ptr);
}
#endif

/* This is the function that does the actual writing of data.  If you are
   not writing to a standard C stream, you should create a replacement
   write_data function and use it at run time with png_set_write_fn(), rather
   than changing the library. */
#ifndef USE_FAR_KEYWORD
static void
png_default_write_data(png_structp png_ptr, png_bytep data, png_size_t length)
{
   png_uint_32 check;

   check = fwrite(data, 1, length, (FILE *)(png_ptr->io_ptr));
   if (check != length)
   {
      png_error(png_ptr, "Write Error");
   }
}
#else
/* this is the model-independent version. Since the standard I/O library
   can't handle far buffers in the medium and small models, we have to copy
   the data.
*/

#define NEAR_BUF_SIZE 1024
#define MIN(a,b) (a <= b ? a : b)



static void
png_default_write_data(png_structp png_ptr, png_bytep data, png_size_t length)
{
   png_uint_32 check;
   png_byte *near_data;  /* Needs to be "png_byte *" instead of "png_bytep" */
   FILE *io_ptr;

   /* Check if data really is near. If so, use usual code. */
   near_data = (png_byte *)CVT_PTR_NOCHECK(data);
   io_ptr = (FILE *)CVT_PTR(png_ptr->io_ptr);
   if ((png_bytep)near_data == data)
   {
      check = fwrite(near_data, 1, length, io_ptr);
   }
   else
   {
      png_byte buf[NEAR_BUF_SIZE];
      png_size_t written, remaining, err;
      check = 0;
      remaining = length;
      do
      {
         written = MIN(NEAR_BUF_SIZE, remaining);
         png_memcpy(buf, data, written); /* copy far buffer to near buffer */
         err = fwrite(buf, 1, written, io_ptr);
         if (err != written)
            break;
         else
            check += err;
         data += written;
         remaining -= written;
      }
      while (remaining != 0);
   }
   if (check != length)
   {
      png_error(png_ptr, "Write Error");
   }
}

#endif /* USE_FAR_KEYWORD */



/* This function is called when there is a warning, but the library thinks
 * it can continue anyway.  Replacement functions don't have to do anything
 * here if you don't want to.  In the default configuration, png_ptr is
 * not used, but it is passed in case it may be useful.
 */
static void
png_default_warning(png_structp png_ptr, png_const_charp message)
{
   PNG_CONST char *name = "UNKNOWN (ERROR!)";
   if (png_ptr != NULL && png_ptr->error_ptr != NULL)
      name = png_ptr->error_ptr;
   fprintf(STDERR, "%s: libpng warning: %s\n", name, message);
}

/* This is the default error handling function.  Note that replacements for
 * this function MUST NOT RETURN, or the program will likely crash.  This
 * function is used by default, or if the program supplies NULL for the
 * error function pointer in png_set_error_fn().
 */
static void
png_default_error(png_structp png_ptr, png_const_charp message)
{
   png_default_warning(png_ptr, message);
   /* We can return because png_error calls the default handler, which is
    * actually OK in this case. */
}
#endif /* PNG_NO_STDIO */
/* END of code to validate stdio-free compilation */



/* START of code to validate memory allocation and deallocation */
#ifdef PNG_USER_MEM_SUPPORTED

/* Allocate memory.  For reasonable files, size should never exceed
   64K.  However, zlib may allocate more then 64K if you don't tell
   it not to.  See zconf.h and png.h for more information.  zlib does
   need to allocate exactly 64K, so whatever you call here must
   have the ability to do that.

   This piece of code can be compiled to validate max 64K allocations
   by setting MAXSEG_64K in zlib zconf.h *or* PNG_MAX_MALLOC_64K. */
typedef struct memory_information {
   png_uint_32                    size;
   png_voidp                 pointer;
   struct memory_information FAR *next;
} memory_information;
typedef memory_information FAR *memory_infop;

static memory_infop pinformation = NULL;
static int current_allocation = 0;
static int maximum_allocation = 0;

extern PNG_EXPORT(png_voidp,png_debug_malloc) PNGARG((png_structp png_ptr,
   png_uint_32 size));
extern PNG_EXPORT(void,png_debug_free) PNGARG((png_structp png_ptr,
   png_voidp ptr));



png_voidp
png_debug_malloc(png_structp png_ptr, png_uint_32 size) {

   /* png_malloc has already tested for NULL; png_create_struct calls
      png_debug_malloc directly, with png_ptr == NULL which is OK */

   if (size == 0)
      return (png_voidp)(NULL);

   /* This calls the library allocator twice, once to get the requested
      buffer and once to get a new free list entry. */
   {
      memory_infop pinfo = png_malloc_default(png_ptr, sizeof *pinfo);
      pinfo->size = size;
      current_allocation += size;
      if (current_allocation > maximum_allocation)
         maximum_allocation = current_allocation;
      pinfo->pointer = png_malloc_default(png_ptr, size);
      pinfo->next = pinformation;
      pinformation = pinfo;
      /* Make sure the caller isn't assuming zeroed memory. */
      png_memset(pinfo->pointer, 0xdd, pinfo->size);
      return (png_voidp)(pinfo->pointer);
   }
}



/* Free a pointer.  It is removed from the list at the same time. */
void
png_debug_free(png_structp png_ptr, png_voidp ptr)
{
   if (png_ptr == NULL)
      fprintf(STDERR, "NULL pointer to png_debug_free.\n");
   if (ptr == 0) {
#if 0 /* This happens all the time. */
      fprintf(STDERR, "WARNING: freeing NULL pointer\n");
#endif
      return;
   }

   /* Unlink the element from the list. */
   {
      memory_infop FAR *ppinfo = &pinformation;
      for (;;) {
         memory_infop pinfo = *ppinfo;
         if (pinfo->pointer == ptr) {
            *ppinfo = pinfo->next;
            current_allocation -= pinfo->size;
            if (current_allocation < 0)
               fprintf(STDERR, "Duplicate free of memory\n");
            /* We must free the list element too, but first kill
               the memory that is to be freed. */
            memset(ptr, 0x55, pinfo->size);
            png_free_default(png_ptr, pinfo);
            break;
         }
         if (pinfo->next == NULL) {
            fprintf(STDERR, "Pointer %x not found\n", ptr);
            break;
         }
         ppinfo = &pinfo->next;
      }
   }

   /* Finally free the data. */
   png_free_default(png_ptr, ptr);
}
#endif /* PNG_USER_MEM_SUPPORTED */
/* END of code to test memory allocation/deallocation */


void WriteBlock(FILE *fpout, double floor, double pix, int x, int y, int pw, int ph)
{
	if (gClip && (pix >= 254/255.0 || pix <= 1/255.0))
		return;
	if (gInvert)
			pix = 1-pix;

	double tw = gMapWidth/gPicWidth;
	double th = tw * (gPicHeight/gPicWidth);
	double	xrad=pw*tw/2.0,
					yrad=ph*th/2.0;
	double	ztall;
	double xo = (x*tw + pw*tw/2.0) - (gMapWidth/2);
	double yo = -((y*th + ph*th/2.0) - (gMapWidth/2));
	double zo;
	if (gTall)
			ztall = gTall;
	if (gLandscape) {
			ztall = (pix-floor)*gMapHeight;
			zo = gFloorOfst+floor*gMapHeight;
	}
	else {
		zo = pix*gMapHeight+gFloorOfst;
		ztall = gTall? gTall : gMapHeight/256;
	}
	fprintf(fpout, "box\n");
	fprintf(fpout, "position %6.3f %6.3f %6.3f\n", xo,yo,zo);
	fprintf(fpout, "rotation 0\n");
	fprintf(fpout, "size %6.3f %6.3f %6.3f\n",xrad,yrad,ztall);
	fprintf(fpout, "end\n\n");
}



/* Convert one file */
int
convert_one_file(PNG_CONST char *inname, PNG_CONST char *outname)
{
   static FILE *fpin, *fpout;  /* "static" prevents setjmp corruption */
   png_structp read_ptr /*, write_ptr*/ ;
   png_infop read_info_ptr /*, write_info_ptr*/ , end_info_ptr;
   png_bytep row_buf;
   png_uint_32 row, col;
	 ImageRef		img;
   //png_uint_32 y;
   //int num_pass, pass;
   int bit_depth, color_type;
#ifdef USE_FAR_KEYWORD
   jmp_buf jmpbuf;
#endif

   //char inbuf[256], outbuf[256];

   row_buf = (png_bytep)NULL;

   if ((fpin = fopen(inname, "rb")) == NULL)
      {
      fprintf(STDERR, "Could not find input file %s\n", inname);
      return (1);
      }

   if ((fpout = fopen(outname, "w")) == NULL)
      {
      fprintf(STDERR, "Could not open output file %s\n", outname);
      fclose(fpin);
      return (1);
      }

   png_debug(0, "Allocating read and write structures\n");
#ifdef PNG_USER_MEM_SUPPORTED
   read_ptr = png_create_read_struct_2(
      PNG_LIBPNG_VER_STRING, (png_voidp)NULL,
      (png_error_ptr)NULL, (png_error_ptr)NULL, (png_voidp)NULL,
      (png_malloc_ptr)png_debug_malloc, (png_free_ptr)png_debug_free);
#else
   read_ptr = png_create_read_struct(
      PNG_LIBPNG_VER_STRING, (png_voidp)NULL,
      (png_error_ptr)NULL, (png_error_ptr)NULL);
#endif
#if defined(PNG_NO_STDIO)
   png_set_error_fn(read_ptr, (png_voidp)inname, png_default_error,
                    png_default_warning);
#endif

   png_debug(0, "Allocating read_info, write_info and end_info structures\n");
   read_info_ptr = png_create_info_struct(read_ptr);
   end_info_ptr = png_create_info_struct(read_ptr);

   png_debug(0, "Setting jmpbuf for read struct\n");
#ifdef USE_FAR_KEYWORD
   if (setjmp(jmpbuf))
#else
   if (setjmp(read_ptr->jmpbuf))
#endif
      {
      fprintf(STDERR, "%s -> %s: libpng read error\n", inname, outname);
      png_destroy_read_struct(&read_ptr, &read_info_ptr, &end_info_ptr);
//       png_destroy_write_struct(&write_ptr, &write_info_ptr);
      fclose(fpin);
      fclose(fpout);
      return (1);
      }
#ifdef USE_FAR_KEYWORD
   png_memcpy(read_ptr->jmpbuf,jmpbuf,sizeof(jmp_buf));
#endif


   png_debug(0, "Initializing input and output streams\n");
#if !defined(PNG_NO_STDIO)
   png_init_io(read_ptr, fpin);

#else
   png_set_read_fn(read_ptr, (png_voidp)fpin, png_default_read_data);
#endif
   png_set_read_status_fn(read_ptr, NULL);

   png_debug(0, "Reading info struct\n");
   png_read_info(read_ptr, read_info_ptr);

   png_debug(0, "Transferring info struct\n");
      {
      int interlace_type, compression_type, filter_type;

      if (png_get_IHDR(read_ptr, read_info_ptr, &img.width, &img.height, &bit_depth,
                       &color_type, &interlace_type, &compression_type, &filter_type))
         {
         if( color_type != PNG_COLOR_TYPE_RGB_ALPHA
             &&
             color_type != PNG_COLOR_TYPE_RGB
             &&
             color_type != PNG_COLOR_TYPE_GRAY )
            {
					 fprintf(stderr,  "Input file color type must be RGBA, RGB or P8.GRAY" );
					 exit(1);
            }
         }
      }
#if defined(PNG_READ_bKGD_SUPPORTED) && defined(PNG_WRITE_bKGD_SUPPORTED)
      {
      png_color_16p background;

      if (png_get_bKGD(read_ptr, read_info_ptr, &background))
         {
//          png_set_bKGD(write_ptr, write_info_ptr, background);
         }
      }
#endif
#if defined(PNG_READ_cHRM_SUPPORTED) && defined(PNG_WRITE_cHRM_SUPPORTED)
      {
      double white_x, white_y, red_x, red_y, green_x, green_y, blue_x, blue_y;

      if (png_get_cHRM(read_ptr, read_info_ptr, &white_x, &white_y, &red_x,
                       &red_y, &green_x, &green_y, &blue_x, &blue_y))
         {
//          png_set_cHRM(write_ptr, write_info_ptr, white_x, white_y, red_x,
//             red_y, green_x, green_y, blue_x, blue_y);
         }
      }
#endif
#if defined(PNG_READ_gAMA_SUPPORTED) && defined(PNG_WRITE_gAMA_SUPPORTED)
      {
      double gamma;

      if (png_get_gAMA(read_ptr, read_info_ptr, &gamma))
         {
//          png_set_gAMA(write_ptr, write_info_ptr, gamma);
         }
      }
#endif
#if defined(PNG_READ_sRGB_SUPPORTED) && defined(PNG_WRITE_sRGB_SUPPORTED)
      {
      int intent;

      if (png_get_sRGB(read_ptr, read_info_ptr, &intent))
         {
//          png_set_sRGB(write_ptr, write_info_ptr, intent);
         }
      }
#endif
#if defined(PNG_READ_hIST_SUPPORTED) && defined(PNG_WRITE_hIST_SUPPORTED)
      {
      png_uint_16p hist;

      if (png_get_hIST(read_ptr, read_info_ptr, &hist))
         {
//          png_set_hIST(write_ptr, write_info_ptr, hist);
         }
      }
#endif
#if defined(PNG_READ_oFFs_SUPPORTED) && defined(PNG_WRITE_oFFs_SUPPORTED)
      {
      png_int_32 offset_x, offset_y;
      int unit_type;

      if (png_get_oFFs(read_ptr, read_info_ptr,&offset_x,&offset_y,&unit_type))
         {
//          png_set_oFFs(write_ptr, write_info_ptr, offset_x, offset_y, unit_type);
         }
      }
#endif
#if defined(PNG_READ_pCAL_SUPPORTED) && defined(PNG_WRITE_pCAL_SUPPORTED)
      {
      png_charp purpose, units;
      png_charpp params;
      png_int_32 X0, X1;
      int type, nparams;

      if (png_get_pCAL(read_ptr, read_info_ptr, &purpose, &X0, &X1, &type,
                       &nparams, &units, &params))
         {
//          png_set_pCAL(write_ptr, write_info_ptr, purpose, X0, X1, type,
//             nparams, units, params);
         }
      }
#endif
#if defined(PNG_READ_pHYs_SUPPORTED) && defined(PNG_WRITE_pHYs_SUPPORTED)
      {
      png_uint_32 res_x, res_y;
      int unit_type;

      if (png_get_pHYs(read_ptr, read_info_ptr, &res_x, &res_y, &unit_type))
         {
//          png_set_pHYs(write_ptr, write_info_ptr, res_x, res_y, unit_type);
         }
      }
#endif
      {
      png_colorp palette;
      int num_palette;

      if (png_get_PLTE(read_ptr, read_info_ptr, &palette, &num_palette))
         {
//          png_set_PLTE(write_ptr, write_info_ptr, palette, num_palette);
         }
      }
#if defined(PNG_READ_sBIT_SUPPORTED) && defined(PNG_WRITE_sBIT_SUPPORTED)
      {
      png_color_8p sig_bit;

      if (png_get_sBIT(read_ptr, read_info_ptr, &sig_bit))
         {
//          png_set_sBIT(write_ptr, write_info_ptr, sig_bit);
         }
      }
#endif
#if (defined(PNG_READ_tEXt_SUPPORTED) && defined(PNG_WRITE_tEXt_SUPPORTED)) || \
    (defined(PNG_READ_zTXt_SUPPORTED) && defined(PNG_WRITE_zTXt_SUPPORTED))
      {
      png_textp text_ptr;
      int num_text;

      if (png_get_text(read_ptr, read_info_ptr, &text_ptr, &num_text) > 0)
         {
         png_debug1(0, "Handling %d tEXt/zTXt chunks\n", num_text);
//          png_set_text(write_ptr, write_info_ptr, text_ptr, num_text);
         }
      }
#endif
#if defined(PNG_READ_tIME_SUPPORTED) && defined(PNG_WRITE_tIME_SUPPORTED)
      {
      png_timep mod_time;

      if (png_get_tIME(read_ptr, read_info_ptr, &mod_time))
         {
//          png_set_tIME(write_ptr, write_info_ptr, mod_time);
#if defined(PNG_TIME_RFC1123_SUPPORTED)
         /* we have to use png_strcpy instead of "=" because the string
            pointed to by png_convert_to_rfc1123() gets free'ed before
            we use it */
         png_strcpy(tIME_string,png_convert_to_rfc1123(read_ptr, mod_time));
         tIME_chunk_present++;
#endif /* PNG_TIME_RFC1123_SUPPORTED */
         }
      }
#endif
#if defined(PNG_READ_tRNS_SUPPORTED) && defined(PNG_WRITE_tRNS_SUPPORTED)
      {
      png_bytep trans;
      int num_trans;
      png_color_16p trans_values;

      if (png_get_tRNS(read_ptr, read_info_ptr, &trans, &num_trans,
                       &trans_values))
         {
//          png_set_tRNS(write_ptr, write_info_ptr, trans, num_trans,
//             trans_values);
         }
      }
#endif

   png_debug(0, "\nWriting info struct\n");
//    png_write_info(write_ptr, write_info_ptr);

   png_debug(0, "\nAllocating row buffer \n");
   row_buf = (png_bytep)png_malloc(read_ptr,
      png_get_rowbytes(read_ptr, read_info_ptr));
   if (row_buf == NULL)
      {
      fprintf(STDERR, "No memory to allocate row buffer\n");
      png_destroy_read_struct(&read_ptr, &read_info_ptr, (png_infopp)NULL);
//       png_destroy_write_struct(&write_ptr, &write_info_ptr);
      fclose(fpin);
      fclose(fpout);
      return (1);
      }
   png_debug(0, "Writing row data\n");

   // allocate all the row buffers
   img.image = new png_bytep[ img.height ];
   img.hmap  = new png_bytep[ img.height ];
   for( row = 0 ; row < img.height ; row++ )
   {
		int rowBytes = png_get_rowbytes(read_ptr, read_info_ptr);
		 img.image[row] = (png_bytep)png_malloc(read_ptr, rowBytes);
      img.hmap[row] = (png_bytep)png_malloc(read_ptr, rowBytes);
			memset(img.hmap[row],0,rowBytes);
   }
   png_read_image( read_ptr, img.image );

	 switch( color_type ) {
   case PNG_COLOR_TYPE_RGB_ALPHA:	img.pixelSize = 4;	break;
   case PNG_COLOR_TYPE_RGB:				img.pixelSize = 3;	break;
	 default:	img.pixelSize = 1;	break;
	 }

	 gPicWidth = img.width;
	 gPicHeight = img.height;

		char	histoGram[256];
		memset(histoGram,0,256);
		for( row = 0 ; row < img.height ; row++ )
			 for( col = 0 ; col < img.width ; col++ )
					histoGram[img.image[row][col*img.pixelSize]] = 1;

		// for each non-zero value for which we have at least one pixel
	  int lastMatch = 0;
		for (int match = 1; match <= 255; ++match)
		{
			if (histoGram[match] == 0)
				continue;

			// decompose this level into blocks
		 for( row = 0 ; row < img.height ; row++ )
		 {
			 for( col = 0 ; col < img.width ; col++ )
				{
						int pix = img.image[row][col*img.pixelSize];
						int flr = img.hmap[row][col*img.pixelSize];
						if (flr < match && pix >= match)
						{
							int	pw,ph;
							MeasureLargestBlock(col,row, match, &img, &pw, &ph);
							if (GetLowestPixel(col, row, pw, ph, &img) == match)
							{
								flr = FindHighest(col, row, pw, ph, &img);
								WriteBlock(fpout, flr/255.0,match/255.0,col,row,pw, ph);
								// Raise floor in this area so we don't rewrite blocks
								MarkBlock(col, row, pw, ph, match, &img);
							}
						}
				}
		 }
		 lastMatch = match;
		}

   png_debug(0, "Reading and writing end_info data\n");
   png_read_end(read_ptr, end_info_ptr);


   png_debug(0, "Destroying data structs\n");
   delete img.image;
	 delete img.hmap;
   png_free(read_ptr, row_buf);
   png_destroy_read_struct(&read_ptr, &read_info_ptr, &end_info_ptr);
//    png_destroy_write_struct(&write_ptr, &write_info_ptr);

   fclose(fpin);
   fclose(fpout);

   return (0);
}


void MeasureLargestBlock(int col, int row, int match, ImageRef *img, int *pw, int *ph)
{
	int	largestWidth = MeasureWidth(col,row,img,match);
	int	minH = MeasureHeight(col,row,img,match);
	int	greatestArea = 1*minH;
	int	sx = 1;
	int sy = minH;
	for (int x = 1; x < largestWidth; ++x) {
		int h = MeasureHeight(col+x,row,img,match);
		if (h > minH)
			h = minH;
		else if (h < minH)
			minH = h;
		if (h*(x+1) > greatestArea) {
			greatestArea = h*(x+1);
			sy = h;
			sx = x+1;
		}
	}
	int largestHeight = MeasureHeight(col,row,img,match);
	int minW = MeasureWidth(col,row,img,match);
	for (int y = 1; y < largestHeight; ++y) {
		int w = MeasureWidth(col,row+y,img,match);
		if (w > minW)
			w = minW;
		else if (w < minW)
			minW = w;
		if (w*(y+1) > greatestArea) {
			greatestArea = w*(y+1);
			sy = y+1;
			sx = w;
		}
	}
	*ph = sy;
	*pw = sx;
}

int MeasureWidth(int col, int row, ImageRef *img, int match)
{
	int pw = 1;
	int ox = col;
	int oy = row;
	// image[row][col*skip] = 0;
	while (col+pw < (int) img->width-1 && img->image[row][(col+pw)*img->pixelSize] >= match && img->hmap[row][(col+pw)*img->pixelSize] < match)
	{
			++pw;
	}
	return pw;
}


int MeasureHeight(int col, int row, ImageRef *img, int match)
{
	int ph = 1;
	int ox = col;
	int oy = row;
	// image[row][col*skip] = 0;
	while (row+ph < (int) img->height-1 && img->image[row+ph][col*img->pixelSize] >= match && img->hmap[row+ph][col*img->pixelSize] < match)
	{
			++ph;
	}
	return ph;
}

int FindHighest(int col, int row, int pw, int ph, ImageRef *img)
{
	int	high = 0;
	for (int y = 0; y < ph; ++y)
	{
		for (int x = 0; x < pw; ++x)
		{
			if (img->hmap[row+y][(col+x)*img->pixelSize] > high)
				high = img->hmap[row+y][(col+x)*img->pixelSize];
		}
	}
	return high;
}

int GetLowestPixel(int col, int row, int pw, int ph, ImageRef *img)
{
	int	low = 256;
	for (int y = 0; y < ph; ++y)
	{
		for (int x = 0; x < pw; ++x)
		{
			if (img->image[row+y][(col+x)*img->pixelSize] < low)
				low = img->image[row+y][(col+x)*img->pixelSize];
		}
	}
	return low;
}



void MarkBlock(int col, int row, int pw, int ph, int match, ImageRef *img)
{
	for (int y = 0; y < ph; ++y)
	{
		for (int x = 0; x < pw; ++x)
		{
			img->hmap[row+y][(col+x)*img->pixelSize] = match;
		}
	}
}




void SyntaxError()
{
  fprintf(STDERR, "\nUsage: png2bzw <in.png> <out.bzw>\n\nOptions:\n");
	fprintf(STDERR, "\t-i Invert Heights\n");
	fprintf(STDERR, "\t-l Landscape (all blocks go down to floor)\n");
	fprintf(STDERR, "\t-f# Floor Offset\n");
	fprintf(STDERR, "\t-t# Tile Height (defaults to H/256)\n");
	fprintf(STDERR, "\t-h# Overall Height\n");
	fprintf(STDERR, "\t-h# Overall Width\n");
	fprintf(STDERR, "\t-v Verbose\n");
	fprintf(STDERR, "\t-m Multiple Input Files\n");

 	exit(1);
}

/* input and output filenames */
// #ifdef RISCOS
// PNG_CONST char *inname = "pngtest/png";
// PNG_CONST char *outname = "pngout/png";
// #else
static PNG_CONST char *inname = NULL;
static PNG_CONST char *outname = NULL;
static char outbuf[256];

// #endif

char *genOutputName(const char *instr)
{
	strcpy(outbuf, instr);
	char *p = strchr(outbuf,'.');
	char isUpper = 0;
	if (p != NULL) {
		isUpper = (p[1] == 'P');
		*p = '\0';
	}
	strcat(outbuf,isUpper? ".bzw" : ".bzw");
	return outbuf;
}

int
main(int argc, char *argv[])
{
   int ierror = 0, firstFile=0,  multiple=0;

   /* Do some consistency checking on the memory allocation settings, I'm
      not sure this matters, but it is nice to know, the first of these
      tests should be impossible because of the way the macros are set
      in pngconf.h */
#if defined(MAXSEG_64K) && !defined(PNG_MAX_MALLOC_64K)
      fprintf(STDERR, " NOTE: Zlib compiled for max 64k, libpng not\n");
#endif
   /* I think the following can happen. */
#if !defined(MAXSEG_64K) && defined(PNG_MAX_MALLOC_64K)
      fprintf(STDERR, " NOTE: libpng compiled for max 64k, zlib not\n");
#endif

	int i;

		// rewrote parser to simplify option handling - jbum
	 for (i = 1; i < argc; ++i)
	 {
		 if (argv[i][0] == '-') {
				for (char *p = &argv[i][1]; *p; ++p)
				{
					switch(tolower(*p))
					{
					case 'v':
						gVerbose = 1;
						break;
					case 'i':
						gInvert= 1;
						break;
					case 'l':
						gLandscape = 1;
						break;
					case 'c':	// clip very dark or light pixels
						gClip = 1;
						break;
					case 't':
						{
							char *arg;
							if (p[1])
								arg = p+1;
							else {
								arg = argv[i+1];
								++i;
							}
							gTall= atof(arg);
							p = &argv[i][strlen(argv[i])-1];
						}
						break;
					case 'f':
						{
							char *arg;
							if (p[1])
								arg = p+1;
							else {
								arg = argv[i+1];
								++i;
							}
							gFloorOfst= atof(arg);
							p = &argv[i][strlen(argv[i])-1];
						}
						break;
					case 'h':
						{
							char *arg;
							if (p[1])
								arg = p+1;
							else {
								arg = argv[i+1];
								++i;
							}
							gMapHeight= atof(arg);
							p = &argv[i][strlen(argv[i])-1];
						}
						break;
					case 'w':
						{
							char *arg;
							if (p[1])
								arg = p+1;
							else {
								arg = argv[i+1];
								++i;
							}
							gMapWidth= atof(arg);
							p = &argv[i][strlen(argv[i])-1];
						}
						break;
					case 'm':
						multiple = 1;	// only strictly needed if two input files
													// otherwise, we figure it out
						break;
					default:
						SyntaxError();
					}
			 }

		 }
		 else {
			 if (firstFile == 0)
				 firstFile = i;
			 if (outname)
			 	 multiple = 1;
			 else if (inname)
         outname = argv[i];
			 else
         inname = argv[i];
		 }

	 }

	 if (inname == NULL)
	 	SyntaxError();

	 // Use input name to generate output name
	 if (outname == NULL) {
		 outname = genOutputName(inname);
	 }


	 int kerror = 0;
#ifdef PNG_USER_MEM_SUPPORTED
	 int allocation_now = current_allocation;
#endif
	if (multiple) {
			for (int i = firstFile; i < argc && !kerror; ++i) {
				inname = argv[i];
				outname = genOutputName(inname);
				if (gVerbose)
					fprintf(STDERR, "Converting %s -> %s\n",inname, outname);
				kerror = convert_one_file(inname, outname);
			}
	}
	else {
	 if (gVerbose)
			fprintf(STDERR, "Converting %s -> %s\n",inname, outname);
	 kerror = convert_one_file(inname, outname);
	}

  return (int)(rc != 0 || kerror != 0);
}

